"""

Finding a basis for the highest weight vectors of shape (5,5,5,4) in Sym^19 otimes^3 (CC^4)^*.

Multi-prolongation patterns
[5, 4, 4, 4, 1, 1]
[5, 4, 4, 3, 2, 1]
[5, 4, 4, 2, 2, 2]
[5, 4, 3, 3, 3, 1]
[5, 4, 3, 3, 2, 2]
[5, 3, 3, 3, 3, 2]
[4, 4, 4, 4, 2, 1]
[4, 4, 4, 3, 3, 1]
[4, 4, 4, 3, 2, 2]
[4, 4, 3, 3, 3, 2]
[4, 3, 3, 3, 3, 3]

This code should work in PyPy.

References
[1] Hauenstein, Ikenemeyer, Landsberg. Equations for lower bounds on border rank.
[2] Christandl, Zuiddam. The border support rank of two-by-two matrix multiplication is seven.

Dependencies
- scoop, for distributed computing. Available at http://pyscoop.org
- Sage, for computing Kronecker coefficients and kernels. Available at http://www.sagemath.org
- blist Python package (but uses dict if not available, for example with pypy)
- numpy Python package, for rank computations.
"""
from scoop import futures
from straighten_GL import straighten3h

from itertools import product, izip_longest, chain
import time
import random
import copy
import unittest
from timeit import default_timer as timer
import subprocess
try:
    from blist import sorteddict # this does not work with pypy
    best_dict = sorteddict
    print "Info: using sorteddict from blist"
except:
    print "Info: sorteddict not available, using standard dict"
    best_dict = dict

def evaluate_job(evaluator):
    return evaluator.evaluate()

def evaluate_straighten_job(evaluator):
    return evaluator.evaluate_straighten()


class Zuiddam(object):
    """
    Methods for constructing a basis of the highest weight vectors of shape (5,5,5,4) in Sym^19 otimes^3 (CC^4)^*.
    """

    def __init__(self, n = 5, pr=None, mod=32):
        self.n = n
        self.pr = pr
        self.mod = mod

    @classmethod
    def grouper(cls,iterable, n, fillvalue=None):
        "Collect data into fixed-length chunks or blocks"
        # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
        args = [iter(iterable)] * n
        return izip_longest(fillvalue=fillvalue, *args)

    @classmethod
    def has_zero_pattern(cls,T1,T2,T3):
        """
        True if the tableaux T1, T2, T3 have a zero pattern: there are two number y and y' such that y and y' occur
        in the same column in T1 and they occur in the same column in T2 and they occur in the same colum in T3.
        The tableaux should be given as lists of columns.
        """
        for c1, c2, c3 in product(T1, T2, T3):
            if len(set(c1).intersection(c2).intersection(c3)) >= 2:
                return True
        return False

    @classmethod
    def random_perms(cls,n=5):
        m = 19
        while True:
            T1 = list(cls.grouper(range(m), 4, 'EmptyA'))
            L = range(m)
            random.shuffle(L)
            L2 = copy.deepcopy(L)
            T2 = list(cls.grouper(L2, 4, 'EmptyB'))
            random.shuffle(L)
            L3 = copy.deepcopy(L)
            T3 = list(cls.grouper(L3, 4, 'EmptyC'))
            if not cls.has_zero_pattern(T1, T2, T3):
                return (L2, L3)

    @classmethod
    def matrix_rank(cls, A):
        """
        Compute the matrix rank of A. We use an auxiliary file rank.py in order
        to use python instead of pypy, in case this file is run with pypy. This
        is needed because we use numpy for matrix rank.
        """
        rank = subprocess.check_output(["python", "rank.py", str(A)])
        return int(rank)

    @classmethod
    def left_kernel_basis(cls, A):
        """
        Compute a left kernel of the integer matrix A. We use an auxiliary file
        left_kernel_basis.sage in order to use Sage. This function dependes on
        Sage.
        """
        kernel = subprocess.check_output(["sage", "left_kernel_basis.sage", str(A)])
        return eval(kernel)

    def construct_basis(self):
        """
        We construct a basis for the higest weight vectors by evaluating in
        randomly perturbed mamu's.
        """
        logfilename = "construct_basis%s.txt" % str(time.time())

        E = Evaluator

        rank_goal = 31
        max_tries = 2*rank_goal #30
        safety_extra = 1

        perturbations = []
        for j in range(rank_goal+safety_extra):
            w1 = [random.randint(-10,10) for i in range(4)]
            w2 = [random.randint(-10,10) for i in range(4)]
            w3 = [random.randint(-10,10) for i in range(4)]
            w4 = [random.randint(-10,10) for i in range(4)]
            w5 = [random.randint(-10,10) for i in range(4)]
            w6 = [random.randint(-10,10) for i in range(4)]
            w7 = [random.randint(-10,10) for i in range(4)]
            w8 = [random.randint(-10,10) for i in range(4)]
            perturbation = [w1, w2, w3, w4, w5, w6, w7, w8]
            perturbations.append(perturbation)
        logfile = open(logfilename,'w',0)
        logfile.write('perturbations: ' + str(perturbations) + '\n')
        logfile.write('prime: ' + str(self.pr) + '\n')
        logfile.write('permutations:\n')

        print 'perturbations:', perturbations
        print 'prime:', self.pr
        print "permutations:"
        rows = []
        permutations = []
        for j in range(max_tries):
            t1, t2 = Zuiddam.random_perms(n=self.n)
            permutations.append((t1,t2))
            logfile.write(str(j) + ', ' + str((t1, t2)) + '\n')
            print j, (t1, t2)

            row = []
            for perturbation in perturbations:

                inputs = []
                for res in range(self.mod):
                    inputs.append(E(t1=t1, t2=t2, pr=self.pr, res=res, mod=self.mod, mamu_perturb=perturbation))
                outputs = futures.map(evaluate_job, inputs)
                row.append(outputs)
            rows.append(row)
        logfile.close()

        basis = []
        matrix = []
        rank = 0

        for i,row in enumerate(rows):
            logfile = open(logfilename, 'a',0)
            logfile.write('\npermutation ' + str(i) + '\n')
            logfile.close()
            print '\npermutation', i
            new_matrix = copy.deepcopy(matrix)
            row_completed = []
            #for a in futures.as_completed(row):
            for outputs in row:
                entry = sum(list(outputs))
                #entry = a.result()
                row_completed.append(entry)
                print entry,
                logfile = open(logfilename, 'a',0)
                logfile.write(str(entry) + ' ')
                logfile.close()
            new_matrix.append(row_completed)
            new_rank = Zuiddam.matrix_rank(new_matrix)
            if new_rank > rank:
                matrix = new_matrix
                rank = new_rank
                basis.append(permutations[i])
                logfile = open(logfilename, 'a',0)
                logfile.write("rank increased to %d" % rank)
                logfile.close
                print 'rank increased to', rank
            if rank == rank_goal:
                break
        logfile = open(logfilename, 'a',0)
        logfile.write(str(matrix))
        logfile.write(str(basis))
        logfile.write(str(rank))
        logfile.close
        print matrix
        print basis
        print rank
        outfile = open("basis%s.txt" % (str(time.time())), 'w')
        outfile.write(str(basis) + '\n')
        outfile.close()
        outfile = open("basis_matrix%s.txt" % (str(time.time())), 'w')
        outfile.write(str(matrix) + '\n')
        outfile.close()
        return basis

    def construct_ideal_basis(self, basis, r=6):
        """
        We construct a basis for the rank at most 6 ideal.
        """
        logfilename = "construct_ideal_basis%s.txt" % str(time.time())

        E = Evaluator
        rank_goal = 31

        tensors = []
        for k in range(rank_goal):
            tensor = [[[random.randint(-10,10) for i in range(4)] for p in range(3)] for j in range(r)]
            tensors.append(tensor)

        logfile = open(logfilename,'w',0)
        logfile.write('tensors: ' + str(tensors) + '\n')
        logfile.write('prime: ' + str(self.pr) + '\n')
        logfile.write('basis:\n')

        print 'tensors:', tensors
        print 'prime:', self.pr
        print "basis:"
        rows = []

        j = 0
        for t1,t2 in basis:
            logfile.write(str(j) + ', ' + str((t1, t2)) + '\n')
            print j, (t1, t2)

            row = []
            for tensor in tensors:

                inputs = []
                for res in range(self.mod):
                    inputs.append(E(t1=t1, t2=t2, pr=self.pr, res=res, mod=self.mod, tensor=tensor))
                outputs = futures.map(evaluate_job, inputs)
                row.append(outputs)
            rows.append(row)
            j+=1
        logfile.close()

        matrix = []

        for i,row in enumerate(rows):
            logfile = open(logfilename, 'a',0)
            logfile.write('\nbasiselt ' + str(i) + '\n')
            logfile.close()
            print '\nbasiselt', i

            row_completed = []
            for outputs in row:
                entry = sum(list(outputs))
                row_completed.append(entry)
                print entry,
                logfile = open(logfilename, 'a',0)
                logfile.write(str(entry) + ' ')
                logfile.close()
            matrix.append(row_completed)

        kernel_basis = Zuiddam.left_kernel_basis(matrix)
        rank = Zuiddam.matrix_rank(matrix)
        logfile = open(logfilename, 'a',0)
        logfile.write(str(matrix) + '\n')
        logfile.write(str(rank) + '\n')
        logfile.write(str(kernel_basis) + '\n')
        logfile.close
        print matrix
        print rank
        print kernel_basis

        outfile = open("ideal_basis%s.txt" % (str(time.time())), 'w')
        outfile.write(str(kernel_basis))
        outfile.close()
        outfile = open("ideal_basis_matrix%s.txt" % (str(time.time())), 'w')
        outfile.write(str(matrix))
        outfile.close()

        return kernel_basis

    def verify_ideal_basis_shelve(self, basis, coeffs_list, r=6, mp_pattern=None):
        """
        We construct a basis for the rank at most 6 ideal.
        """
        logfilename = "verify_ideal_basis%s.txt" % str(time.time())

        import shelve
        import gc

        E = Evaluator
        rank_goal = 31

        logfile = open(logfilename,'w',0)
        logfile.write('prime: ' + str(self.pr) + '\n')
        logfile.write('coeffs: ' + str(coeffs_list) + '\n')
        logfile.write('rank: ' + str(r) + '\n')
        logfile.write('mp_pattern: ' + str(mp_pattern) + '\n')
        logfile.write('basis:\n')

        print 'prime:', self.pr
        print 'coeffs:', coeffs_list
        print 'rank:', r
        print 'mp_pattern:', mp_pattern
        print "basis:"

        outputs_list = []

        j = 0
        for t1,t2 in basis:
            logfile.write(str(j) + ', ' + str((t1, t2)) + '\n')
            print j, (t1, t2)

            inputs = []
            for res in range(self.mod):
                inputs.append(E(t1=t1, t2=t2, pr=self.pr, res=res, mod=self.mod,r=r, mp_pattern=mp_pattern))
            # To prevent memory problems, we do a map_as_completed.
            outputs = futures.map_as_completed(evaluate_straighten_job, inputs)
            outputs_list.append(outputs)
            j+=1
        logfile.close()

        basis_evals = []


        for i,outputs in enumerate(outputs_list):

            logfile = open(logfilename, 'a',0)
            logfile.write('\nbasiselt ' + str(i) + ' ')
            logfile.close()
            print '\nbasiselt', i,

            entry = best_dict()

            for output in outputs:
                #print output
                for k,v in output.iteritems():
                    if k in entry:
                        x = entry[k] + v
                        if x == 0:
                            del entry[k]
                        else:
                            entry[k] = x
                    else:
                        entry[k] = v
                output = None # free memory
            outputs = None # free memory

            # Assuming we have a scratch disk
            entry_shelve = shelve.open("/scratch/shelve%d-%s" % (i,str(time.time())), writeback=False)
            for k,v in entry.iteritems():
                entry_shelve[str(k)] = v
            entry_shelve.sync()
            del entry
            #gc.collect()

            basis_evals.append(entry_shelve)
            print 'done'
            logfile = open(logfilename, 'a',0)
            logfile.write('done\n')
            logfile.close()

        # BUG: this only works if coeffs_list has length 1.
        for coeffs in coeffs_list:
            for i in range(len(basis_evals)):
                for k,v in basis_evals[i].iteritems():
                    running_sum = coeffs[i] * v
                    for j in range(i+1,len(basis_evals)):
                        if k in basis_evals[j]:
                            running_sum += coeffs[j] * basis_evals[j][k]
                            del basis_evals[j][k]
                    if not running_sum == 0:
                        print "verification failed"
                        outfile = open(logfilename, 'a',0)
                        outfile.write("verification failed")
                        outfile.close()
                        return False

        print "verification succeeded"
        outfile = open(logfilename, 'a',0)
        outfile.write("verification succeeded\n")
        outfile.close()

        return True

    def verify_ideal_basis_shelve_single(self, basis, coeffs_list, r=6, mp_pattern=None, nr=0):
        """
        We construct a basis for the rank at most 6 ideal.
        """
        logfilename = "verify_ideal_basis%d-%s.txt" % (nr, str(time.time()))

        import shelve
        #import gc

        E = Evaluator
        rank_goal = 31

        logfile = open(logfilename,'w',0)
        logfile.write('prime: ' + str(self.pr) + '\n')
        logfile.write('coeffs: ' + str(coeffs_list) + '\n')
        logfile.write('rank: ' + str(r) + '\n')
        logfile.write('mp_pattern: ' + str(mp_pattern) + '\n')
        logfile.write('basis:\n')

        print 'prime:', self.pr
        print 'coeffs:', coeffs_list
        print 'rank:', r
        print 'mp_pattern:', mp_pattern
        print "basis:"

        outputs_list = []
        entry_shelve = shelve.open("shelve%d" % nr, writeback=False)

        j = 0
        for t1,t2 in basis:
            if not j == nr:
                j+=1
                continue
            logfile.write(str(j) + ', ' + str((t1, t2)) + '\n')
            print j, (t1, t2)

            inputs = []
            for res in range(self.mod):
                inputs.append(E(t1=t1, t2=t2, pr=self.pr, res=res, mod=self.mod,r=r, mp_pattern=mp_pattern))
            # To prevent memory problems, we do a map_as_completed.
            outputs = futures.map_as_completed(evaluate_straighten_job, inputs)
            outputs_list.append(outputs)
            j+=1
        logfile.close()

        basis_evals = []


        for i,outputs in enumerate(outputs_list):

            logfile = open(logfilename, 'a',0)
            logfile.write('\nbasiselt ' + str(nr) + ' ')
            logfile.close()
            print '\nbasiselt', nr,

            entry = best_dict()

            for output in outputs:
                #print output
                for k,v in output.iteritems():
                    if k in entry:
                        x = entry[k] + v
                        if x == 0:
                            del entry[k]
                        else:
                            entry[k] = x
                    else:
                        entry[k] = v
                output = None # free memory
            outputs = None # free memory

            # Assuming we have a scratch disk
            #entry_shelve = shelve.open("shelve%d" % nr, writeback=False)
            for k,v in entry.iteritems():
                entry_shelve[str(k)] = v
            entry_shelve.close()

            print 'done'
            logfile = open(logfilename, 'a',0)
            logfile.write('done\n')
            logfile.close()

        return True


    def verify_ideal_basis(self, basis, coeffs_list, r=6, mp_pattern=None):
        """
        We construct a basis for the rank at most 6 ideal.
        """
        logfilename = "verify_ideal_basis%s.txt" % str(time.time())

        E = Evaluator
        rank_goal = 31

        logfile = open(logfilename,'w',0)
        logfile.write('prime: ' + str(self.pr) + '\n')
        logfile.write('coeffs: ' + str(coeffs_list) + '\n')
        logfile.write('rank: ' + str(r) + '\n')
        logfile.write('mp_pattern: ' + str(mp_pattern) + '\n')
        logfile.write('basis:\n')

        print 'prime:', self.pr
        print 'coeffs:', coeffs_list
        print 'rank:', r
        print 'mp_pattern:', mp_pattern
        print "basis:"

        outputs_list = []

        j = 0
        for t1,t2 in basis:
            logfile.write(str(j) + ', ' + str((t1, t2)) + '\n')
            print j, (t1, t2)

            inputs = []
            for res in range(self.mod):
                inputs.append(E(t1=t1, t2=t2, pr=self.pr, res=res, mod=self.mod,r=r, mp_pattern=mp_pattern))
            outputs = futures.map_as_completed(evaluate_straighten_job, inputs)
            outputs_list.append(outputs)
            j+=1
        logfile.close()

        basis_evals = []

        for i,outputs in enumerate(outputs_list):
            logfile = open(logfilename, 'a',0)
            logfile.write('\nbasiselt ' + str(i) + ' ')
            logfile.close()
            print '\nbasiselt', i,

            entry = best_dict()
            for output in outputs:
                #print output
                for k,v in output.iteritems():
                    if k in entry:
                        x = entry[k] + v
                        if x == 0:
                            del entry[k]
                        else:
                            entry[k] = x
                    else:
                        entry[k] = v
            basis_evals.append(entry)
            print 'done'
            logfile = open(logfilename, 'a',0)
            logfile.write('done\n')
            logfile.close()

        # BUG: this only works if coeffs_list has length 1.
        for coeffs in coeffs_list:
            for i in range(len(basis_evals)):
                for k,v in basis_evals[i].iteritems():
                    running_sum = coeffs[i] * v
                    for j in range(i+1,len(basis_evals)):
                        if k in basis_evals[j]:
                            running_sum += coeffs[j] * basis_evals[j][k]
                            del basis_evals[j][k]
                    if not running_sum == 0:
                        print "verification failed"
                        outfile = open(logfilename, 'a',0)
                        outfile.write("verification failed")
                        outfile.close()
                        return False

        print "verification succeeded"
        outfile = open(logfilename, 'a',0)
        outfile.write("verification succeeded\n")
        outfile.close()

        return True

    def verify_ideal_basis_symbolically(self, basis, coeffs_list, r=6):
        """
        We verify a basis for the rank at most 6 ideal by evaluating at a
        symbolic rank 6 tensor.
        """
        import sympy
        logfilename = "verify_ideal_basis_symb%s.txt" % str(time.time())

        E = Evaluator
        rank_goal = 31

        logfile = open(logfilename,'w',0)
        logfile.write('prime: ' + str(self.pr) + '\n')
        logfile.write('basis:\n')

        print 'prime:', self.pr
        print 'basis:'

        generic_tensor = [[[sympy.var("a%d%d%d" % (i,j,k)) for i in range(4)] for j in range(3)] for k in range(r - 1)] + [[[0,0,0,1],[0,0,0,1],[0,0,0,1]],] # unipotent trick

        outputs_list = []

        j = 0
        for t1,t2 in basis:
            logfile.write(str(j) + ', ' + str((t1, t2)) + '\n')
            print j, (t1, t2)

            inputs = []
            for res in range(self.mod):
                inputs.append(E(t1=t1, t2=t2, pr=self.pr, res=res, tensor = generic_tensor, mod=self.mod, symbolic=True))
            outputs = futures.map(evaluate_job, inputs)
            outputs_list.append(outputs)
            j+=1
        logfile.close()

        basis_evals = []

        for i,outputs in enumerate(outputs_list):
            logfile = open(logfilename, 'a',0)
            logfile.write('\nbasiselt ' + str(i) + ' ')
            logfile.close()
            print '\nbasiselt', i,

            entry = 0
            for output in outputs:
                entry += output
            basis_evals.append(entry)
            print 'done'
            logfile = open(logfilename, 'a',0)
            logfile.write('done\n')
            logfile.close()

        for coeffs in coeffs_list:
            lin_comb = 0
            for i in range(len(basis_evals)):
                lin_comb += coeffs[i] * basis_evals[i]
            if not sympy.expand(lin_comb) == 0:
                print "verification failed"
                outfile = open(logfilename, 'a',0)
                outfile.write("verification failed")
                outfile.close()
                return False


        print "verification succeeded"
        outfile = open(logfilename, 'a',0)
        outfile.write("verification succeeded\n")
        outfile.close()

        return True

    def evaluate_mamu_param(self, basis, coeffs_list):
        """
        We evaluate a parametrized mamu.
        """
        import sympy
        logfilename = "eval_mamu_param%s.txt" % str(time.time())

        E = Evaluator

        logfile = open(logfilename,'w',0)
        logfile.write('prime: ' + str(self.pr) + '\n')
        logfile.write('basis:\n')

        print 'prime:', self.pr
        print 'basis:'

        param = sympy.var("z")

        outputs_list = []

        j = 0
        for t1,t2 in basis:
            logfile.write(str(j) + ', ' + str((t1, t2)) + '\n')
            print j, (t1, t2)

            inputs = []
            for res in range(self.mod):
                inputs.append(E(t1=t1, t2=t2, pr=self.pr, res=res, mamu_param=param, mod=self.mod, symbolic=True))
            outputs = futures.map(evaluate_job, inputs)
            outputs_list.append(outputs)
            j+=1
        logfile.close()

        basis_evals = []

        for i,outputs in enumerate(outputs_list):
            logfile = open(logfilename, 'a',0)
            logfile.write('\nbasiselt ' + str(i) + ' ')
            logfile.close()
            print '\nbasiselt', i,

            entry = 0
            for output in outputs:
                entry += output
            basis_evals.append(entry)
            print 'done'
            logfile = open(logfilename, 'a',0)
            logfile.write('done\n')
            logfile.close()

        for coeffs in coeffs_list:
            lin_comb = 0
            for i in range(len(basis_evals)):
                lin_comb += coeffs[i] * basis_evals[i]
            exp_lin_comb = sympy.expand(lin_comb)
            print "evaluation:", exp_lin_comb
            outfile = open(logfilename, 'a', 0)
            outfile.write(str(exp_lin_comb))
            outfile.close()
            #if not sympy.expand(lin_comb) == 0:
            #    print "verification failed"
            #    outfile = open(logfilename, 'a',0)
            #    outfile.write("verification failed")
            #    outfile.close()
            #    return False


        #print "verification succeeded"
        #outfile = open(logfilename, 'a',0)
        #outfile.write("verification succeeded\n")
        #outfile.close()

        return True

class TestZuiddam(unittest.TestCase):

    def test_grouper(self):
        self.assertEqual(list(Zuiddam.grouper(range(20), 4)), [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15), (16, 17, 18, 19)])
        self.assertEqual(list(Zuiddam.grouper(range(20), 5)), [(0, 1, 2, 3, 4), (5, 6, 7, 8, 9), (10, 11, 12, 13, 14), (15, 16, 17, 18, 19)])
        L = [1, 8, 11, 3, 0, 2, 10, 14, 15, 13, 12, 17, 9, 6, 19, 4, 5, 16, 7, 18]
        self.assertEqual(list(Zuiddam.grouper(L,4)), [(1, 8, 11, 3), (0, 2, 10, 14), (15, 13, 12, 17), (9, 6, 19, 4), (5, 16, 7, 18)])
        L = range(20)
        random.shuffle(L)
        self.assertEqual(set(L), set(range(20)))
        self.assertEqual(set(chain(*Zuiddam.grouper(L, 4))), set(L))

    def test_has_zero_pattern(self):
        T1 = [[1, 2, 3],[4, 5, 6]]
        T2 = [[1, 2, 4],[3, 5, 6]]
        T3 = [[1, 2, 5],[3, 4, 6]]
        self.assertTrue(Zuiddam.has_zero_pattern(T1, T2, T3))
        T1 = [[1, 2, 3],[4, 5, 6]]
        T2 = [[1, 2, 4],[3, 5, 6]]
        T3 = [[1, 3, 5],[2, 4, 6]]
        self.assertFalse(Zuiddam.has_zero_pattern(T1, T2, T3))

    def test_random_perms(self):
        L2, L3 = Zuiddam.random_perms()
        self.assertEqual(set(L2),set(range(20)))
        self.assertEqual(set(L3),set(range(20)))
        L1 = range(20)
        T1 = Zuiddam.grouper(L1, 4)
        T2 = Zuiddam.grouper(L2, 4)
        T3 = Zuiddam.grouper(L3, 4)
        self.assertFalse(Zuiddam.has_zero_pattern(T1, T2, T3))

    def test_matrix_rank(self):
        A = [[1,0],[0,1]]
        self.assertEqual(Zuiddam.matrix_rank(A), 2)


class TensorData(object):
    # the support of the matrix multiplication tensor
    #
    # ijk  ij jk ki  abc
    # 000  00 00 00  000
    # 001  10 00 01  201
    # 010  01 10 00  120
    # 011  11 10 01  321
    # 100  00 01 10  012
    # 101  10 01 11  213
    # 110  01 11 10  132
    # 111  11 11 11  333

    m = [
        [0, 0, 0],
        [2, 0, 1],
        [1, 2, 0],
        [3, 2, 1],
        [0, 1, 2],
        [2, 1, 3],
        [1, 3, 2],
        [3, 3, 3]]

    u = [[int(i == j) for i in range(4)] for j in range(4)]

    @classmethod
    def matrix_det(cls,l):
        """
        Determinant of a matrix.
        """
        n = len(l)
        if (n > 2):
            i = 1
            t = 0
            sum = 0
            while t <= n - 1:
                d = {}
                t1 = 1
                while t1 <= n - 1:
                    m = 0
                    d[t1] = []
                    while m <= n - 1:
                        if (m == t):
                            u = 0
                        else:
                            d[t1].append(l[t1][m])
                        m += 1
                    t1 += 1
                l1 = [d[x] for x in d]
                sum = sum + i * (l[0][t]) * (cls.matrix_det(l1))
                i = i * (-1)
                t += 1
            return sum
        else:
            return (l[0][0] * l[1][1] - l[0][1] * l[1][0])

    def __init__(self, tensor=None):
        self.precompute_det_mamu()
        if not tensor is None:
            self.tensor = tensor
            self.r = len(tensor)
            self.precompute_det()
            self.precompute_det3()

        from itertools import product
        # Tests.
        assert all(
            (self.det_mamu(a, b, c, d, p) == self._det_mamu(a, b, c, d, p) for a, b, c, d in product(range(8), repeat=4) for p in
             range(3)))
        assert [self.det_mamu(a, b, c, d, 0) for a, b, c, d in
                ((0, 1, 2, 7), (5, 4, 2, 3), (0, 1, 7, 2), (6, 5, 4, 3), (6, 4, 7, 1))] == [-1, 1, 1, 1, 1]
        assert [self.det_mamu(a, b, c, d, 1) for a, b, c, d in
                ((1, 6, 3, 4), (2, 7, 0, 4), (7, 0, 2, 5), (5, 3, 1, 6), (2, 4, 7, 1))] == [-1, 1, 1, 1, 1]
        assert [self.det_mamu(a, b, c, d, 2) for a, b, c, d in
                ((1, 7, 4, 2), (3, 2, 4, 5), (4, 0, 1, 7), (7, 6, 0, 1), (6, 2, 5, 3))] == [1, -1, 1, -1, -1]

    # MaMu related
    def _det_mamu(self, i1, i2, i3, i4, p):
        u = self.u
        m = self.m
        return self.matrix_det([u[m[i1][p]], u[m[i2][p]], u[m[i3][p]], u[m[i4][p]]])

    def precompute_det_mamu(self):
        self.precomputed_det_mamu = [
            [[[[self._det_mamu(a, b, c, d, p) for p in range(3)] for d in range(8)] for c in range(8)] for b in range(8)] for
            a in range(8)]

    def det_mamu(self, i1, i2, i3, i4, p):
        """
        The determinant of the matrix given by the vectors m[i1][p], m[i2][p], m[i3][p], m[i4][p].
        """
        return self.precomputed_det_mamu[i1][i2][i3][i4][p]

    # General
    def _det(self, i1, i2, i3, i4, p):
        tensor = self.tensor
        return self.matrix_det([tensor[i1][p], tensor[i2][p], tensor[i3][p], tensor[i4][p]])

    def precompute_det(self):
        r = self.r
        self.precomputed_det = [
            [[[[self._det(a, b, c, d, p) for p in range(3)] for d in range(r)] for c in range(r)] for b in range(r)] for
            a in range(r)]

    def det(self, i1, i2, i3, i4, p):
        """
        The determinant of the matrix given by the vectors tensor[i1][p], tensor[i2][p], tensor[i3][p], tensor[i4][p].
        """
        return self.precomputed_det[i1][i2][i3][i4][p]

    # size 3

    def _det3(self, i1, i2, i3, p):
        tensor = self.tensor
        return self.matrix_det([tensor[i1][p][0:3], tensor[i2][p][0:3], tensor[i3][p][0:3]])

    def precompute_det3(self):
        r = self.r
        self.precomputed_det3 = [[[[self._det3(a, b, c, p) for p in range(3)] for c in range(r)] for b in range(r)] for a in range(r)]

    def det3(self, i1, i2, i3, p):
        """
        The determinant of the matrix given by the vectors tensor[i1][p], tensor[i2][p], tensor[i3][p].
        """
        return self.precomputed_det3[i1][i2][i3][p]


class Evaluator(object):
    def __init__(self, t1, t2, pr, res=0, mod=0, mamu_param=1, tensor=None, mamu_perturb=None, alt_perturb=False, symbolic=False, r=6, mp_pattern=None):
        """
        Set up an evaluator for evaluating the Ikenmeyer polynomial given by t1 and t2.
        """

        self.t1 = t1
        self.t2 = t2
        self.res = res
        self.mod = mod
        self.pr = pr
        self.mamu_param = mamu_param
        self.tensor = tensor
        self.mamu_perturb = mamu_perturb
        self.alt_perturb = alt_perturb
        self.r = r
        self.mp_pattern = mp_pattern
        if symbolic:
            import sympy

    def evaluate(self):
        from math import ceil

        t1 = self.t1
        t2 = self.t2
        pr = self.pr
        res = self.res
        mod = self.mod

        if not self.tensor is None:
            #print 'tensor'
            tensor = self.tensor
            r = len(tensor)
            T = TensorData(tensor)
            det = T.det
            det3 = T.det3
            clash = self.clash
            level_list = list(self.level_iter(r, s=4))
            level_list_top = list(self.level_iter(r, s=3))

        elif not self.mamu_perturb is None:
            #print 'perturb'
            u = TensorData.u
            m = TensorData.m
            w1, w2, w3, w4, w5, w6, w7, w8 = self.mamu_perturb
            if not self.alt_perturb:
                mamu = [map(lambda i: u[i], mon) for mon in [m[0], m[1], m[2], m[3]]]  # we only take the first 4 summands of MaMu
                mamu_plus_tensor = mamu + [[u[0], w7, w1], [u[2], w8, w2], [u[1], w3, w4], [u[3], w5, w6]]
                T = TensorData(mamu_plus_tensor)
                det = T.det
                det3 = T.det3
                clash = self.clash_mamu_plus_tensor
            else:
                mamu = [map(lambda i: u[i], mon) for mon in [m[4], m[5], m[6], m[7]]]  # we only take the last 4 summands of MaMu
                mamu_plus_tensor = [[u[0], w1, w2], [u[2], w3, w4], [u[1], u[2], w5], [u[3], u[2], w6]] + mamu
                T = TensorData(mamu_plus_tensor)
                det = T.det
                det3 = T.det3
                clash = self.clash_mamu_plus_tensor_alt
            level_list = list(self.level_iter_mamu(s=4))
            level_list_top = list(self.level_iter_mamu(s=3))

        else:
            mamu_param = self.mamu_param
            u = TensorData.u
            m = TensorData.m
            tensor = [map(lambda i: u[i], mon) for mon in m[:-1]] + [[u[3], u[3], [0, 0, 0, mamu_param]],]
            T = TensorData(tensor)
            det = T.det
            det3 = T.det3
            clash = self.clash_mamu
            level_list = list(self.level_iter_mamu(s=4))
            level_list_top = list(self.level_iter_mamu(s=3))

        if mod == 0:
            level_list_top_mod = level_list_top
        else:
            # Careful here, pypy handles division differently than Sage.
            level_list_top_mod = level_list_top[res * int(ceil(float(len(level_list_top))/mod)) : (res + 1) * int(ceil(float(len(level_list_top))/mod))]

        j = (5+5+5+4) * [0, ]
        output = 0

        for j[0], j[1], j[2] in level_list_top_mod:
            if clash(j, 0):
                continue
            for j[3], j[4], j[5], j[6] in level_list:
                if clash(j, 1):
                    continue
                for j[7], j[8], j[9], j[10] in level_list:
                    if clash(j, 2):
                        continue
                    for j[11], j[12], j[13], j[14] in level_list:
                        if clash(j, 3):
                            continue
                        for j[15], j[16], j[17], j[18] in level_list:
                            if clash(j, 4):
                                continue

                            summand = 1

                            if pr:
                                # size 3 det
                                summand = summand  * (det3(j[0 ],    j[1 ],    j[2 ],    0) % pr)
                                summand = (summand * (det3(j[t1[0]], j[t1[1]], j[t1[2]], 1) % pr) % pr)
                                summand = (summand * (det3(j[t2[0]], j[t2[1]], j[t2[2]], 2) % pr) % pr)

                                # size 4 dets
                                for k in range(0,4):
                                    summand = summand  * (det(j[0 + 4 * k + 3],     j[1 + 4 * k + 3],     j[2 + 4 * k + 3],     j[3 + 4 * k + 3],     0) % pr)
                                    summand = (summand * (det(j[t1[0 + 4 * k + 3]], j[t1[1 + 4 * k + 3]], j[t1[2 + 4 * k + 3]], j[t1[3 + 4 * k + 3]], 1) % pr) % pr)
                                    summand = (summand * (det(j[t2[0 + 4 * k + 3]], j[t2[1 + 4 * k + 3]], j[t2[2 + 4 * k + 3]], j[t2[3 + 4 * k + 3]], 2) % pr) % pr)
                                output = (output + summand) % pr
                            else:
                                # size 3 det
                                summand = summand * det3(j[0 ],    j[1 ],    j[2 ],    0)
                                summand = summand * det3(j[t1[0]], j[t1[1]], j[t1[2]], 1)
                                summand = summand * det3(j[t2[0]], j[t2[1]], j[t2[2]], 2)

                                # size 4 dets
                                for k in range(0,4):
                                    summand = summand * det(j[0 + 4 * k + 3],     j[1 + 4 * k + 3],     j[2 + 4 * k + 3],     j[3 + 4 * k + 3],     0)
                                    summand = summand * det(j[t1[0 + 4 * k + 3]], j[t1[1 + 4 * k + 3]], j[t1[2 + 4 * k + 3]], j[t1[3 + 4 * k + 3]], 1)
                                    summand = summand * det(j[t2[0 + 4 * k + 3]], j[t2[1 + 4 * k + 3]], j[t2[2 + 4 * k + 3]], j[t2[3 + 4 * k + 3]], 2)
                                output = output + summand

        return output

    def evaluate_straighten(self):
        from math import ceil

        t1 = self.t1
        t2 = self.t2
        pr = self.pr
        res = self.res
        mod = self.mod
        r = self.r
        mp_pattern = self.mp_pattern


        clash = self.clash
        level_list = list(self.level_iter(r, s=4))
        level_list_top = list(self.level_iter(r, s=3))

        #if mod == 0:
        #    level_list_top_mod = level_list_top
        #else:
        #    # Careful here, pypy handles division differently than Sage.
        #    level_list_top_mod = level_list_top[res * int(ceil(float(len(level_list_top))/mod)) : (res + 1) * int(ceil(float(len(level_list_top))/mod))]

        j = (5+5+5+4) * [0, ]
        output = best_dict()

        parallelization_counter = -1

        for j[0], j[1], j[2] in level_list_top:
            if clash(j, 0):
                continue
            for j[3], j[4], j[5], j[6] in level_list:
                parallelization_counter = (parallelization_counter + 1) % mod
                if not (parallelization_counter % mod == res):
                    continue

                if clash(j, 1):
                    continue
                for j[7], j[8], j[9], j[10] in level_list:
                    if clash(j, 2):
                        continue
                    for j[11], j[12], j[13], j[14] in level_list:
                        if clash(j, 3):
                            continue
                        for j[15], j[16], j[17], j[18] in level_list:
                            if clash(j, 4):
                                continue

                            j_pattern = sorted([j.count(i) for i in set(j)], reverse=True)

                            if not mp_pattern is None and not j_pattern == mp_pattern:
                                continue

                            tab0 = ((j[0 + 4 * 0 + 3], j[1 + 4 * 0 + 3], j[2 + 4 * 0 + 3], j[3 + 4 * 0 + 3]),
                                   (j[0 + 4 * 1 + 3], j[1 + 4 * 1 + 3], j[2 + 4 * 1 + 3], j[3 + 4 * 1 + 3]),
                                   (j[0 + 4 * 2 + 3], j[1 + 4 * 2 + 3], j[2 + 4 * 2 + 3], j[3 + 4 * 2 + 3]),
                                   (j[0 + 4 * 3 + 3], j[1 + 4 * 3 + 3], j[2 + 4 * 3 + 3], j[3 + 4 * 3 + 3]),
                                   (j[0], j[1], j[2]))

                            tab1 = ((j[t1[0 + 4 * 0 + 3]], j[t1[1 + 4 * 0 + 3]], j[t1[2 + 4 * 0 + 3]], j[t1[3 + 4 * 0 + 3]]),
                                   (j[t1[0 + 4 * 1 + 3]], j[t1[1 + 4 * 1 + 3]], j[t1[2 + 4 * 1 + 3]], j[t1[3 + 4 * 1 + 3]]),
                                   (j[t1[0 + 4 * 2 + 3]], j[t1[1 + 4 * 2 + 3]], j[t1[2 + 4 * 2 + 3]], j[t1[3 + 4 * 2 + 3]]),
                                   (j[t1[0 + 4 * 3 + 3]], j[t1[1 + 4 * 3 + 3]], j[t1[2 + 4 * 3 + 3]], j[t1[3 + 4 * 3 + 3]]),
                                   (j[t1[0]], j[t1[1]], j[t1[2]]))

                            tab2 = ((j[t2[0 + 4 * 0 + 3]], j[t2[1 + 4 * 0 + 3]],  j[t2[2 + 4 * 0 + 3]], j[t2[3 + 4 * 0 + 3]]),
                                   (j[t2[0 + 4 * 1 + 3]], j[t2[1 + 4 * 1 + 3]], j[t2[2 + 4 * 1 + 3]], j[t2[3 + 4 * 1 + 3]]),
                                   (j[t2[0 + 4 * 2 + 3]], j[t2[1 + 4 * 2 + 3]], j[t2[2 + 4 * 2 + 3]], j[t2[3 + 4 * 2 + 3]]),
                                   (j[t2[0 + 4 * 3 + 3]], j[t2[1 + 4 * 3 + 3]], j[t2[2 + 4 * 3 + 3]], j[t2[3 + 4 * 3 + 3]]),
                                   (j[t2[0]], j[t2[1]], j[t2[2]]))

                            #h = {}
                            st = straighten3h(tab0,tab1,tab2)
                            for k,v in st.iteritems():
                                if v == 0:
                                    continue
                                else:
                                    k_hash = int(''.join(map(str,sum(k[0] + k[1] + k[2], ()))))
                                    if k_hash in output:
                                        w = output[k_hash] + v
                                        if w == 0:
                                            del output[k_hash]
                                        else:
                                            output[k_hash] = w
                                    else:
                                        output[k_hash] = v

        return output

    #
    # Evaluate mamu
    #

    def clash_mamu(self, j, level):
        t1 = self.t1
        t2 = self.t2
        jt = []

        counts = 10*[0,]
        found = None
        for i in range(3 + 4 * level):
            x = j[i]
            counts[x] += 1
            if counts[x] >= 5:
                if found == None:
                    found = x
                elif not found == x:
                    return True

        #k = 0
        del jt[:]
        for m in (0,1,2):
            i = t1[m]
            if i < 3 + 4 * level:
                if j[i] in jt:
                    # Same index in group of 4.
                    return True
                jt.append(j[i])
        # Specific to mamu.
        for B in ((0, 1), (2, 3), (4, 5), (6, 7)):
            if B[0] in jt and B[1] in jt:
                # Same vector in group of 4.
                return True

        del jt[:]
        for m in (0,1,2):
            i = t2[m]
            if i < 3 + 4 * level:
                if j[i] in jt:
                    # Same index in group of 4.
                    return True
                jt.append(j[i])
        # Specific to mamu.
        for B in ((0, 2), (1, 3), (4, 6), (5, 7)):
            if B[0] in jt and B[1] in jt:
                # Same vector in group of 4.
                return True

        for k in (0,1,2,3):
            del jt[:]
            for m in (0,1,2,3):
                i = t1[m + 4 * k + 3]
                if i < 3 + 4 * level:
                    if j[i] in jt:
                        # Same index in group of 4.
                        return True
                    jt.append(j[i])
            # Specific to mamu.
            for B in ((0, 1), (2, 3), (4, 5), (6, 7)):
                if B[0] in jt and B[1] in jt:
                    # Same vector in group of 4.
                    return True

            del jt[:]
            for m in (0,1,2,3):
                i = t2[m + 4 * k + 3]
                if i < 3 + 4 * level:
                    if j[i] in jt:
                        # Same index in group of 4.
                        return True
                    jt.append(j[i])
            # Specific to mamu.
            for B in ((0, 2), (1, 3), (4, 6), (5, 7)):
                if B[0] in jt and B[1] in jt:
                    # Same vector in group of 4.
                    return True
        return False

    def level_iter_mamu(self, s=4):
        """
        Generates index tuples that are sufficient to evaluate MaMu.

        The indices are such that determinant coming from the first factor is not zero, which can be seen by looking
        at the first tensors leg of each summand in MaMu:

            0 2 1 3 0 2 1 3

        """
        from itertools import product, permutations
        if s == 4:
            for j in permutations(range(4), 4):
                for i in product([j[0], j[0] + 4], [j[1], j[1] + 4], [j[2], j[2] + 4], [j[3], j[3] + 4]):
                    yield (i)
        if s == 3:
            for j in permutations(range(4), 3):
                for i in product([j[0], j[0] + 4], [j[1], j[1] + 4], [j[2], j[2] + 4]):
                    yield (i)


    #
    # Evaluate tensor
    #

    def clash(self, j, level):
        t1 = self.t1
        t2 = self.t2
        jt = []

        counts = 10*[0,]
        found = None
        for i in range(3 + 4 * level):
            x = j[i]
            counts[x] += 1
            if counts[x] >= 5:
                if found == None:
                    found = x
                elif not found == x:
                    return True

        #k = 0
        del jt[:]
        for m in (0,1,2):
            i = t1[m]
            if i < 3 + 4 * level:
                if j[i] in jt:
                    # Same index in group of 4.
                    return True
                jt.append(j[i])

        del jt[:]
        for m in (0,1,2):
            i = t2[m]
            if i < 3 + 4 * level:
                if j[i] in jt:
                    # Same index in group of 4.
                    return True
                jt.append(j[i])

        for k in (0,1,2,3):
            del jt[:]
            for m in (0,1,2,3):
                i = t1[m + 4 * k + 3]
                if i < 3 + 4 * level:
                    if j[i] in jt:
                        # Same index in group of 4.
                        return True
                    jt.append(j[i])

            del jt[:]
            for m in (0,1,2,3):
                i = t2[m + 4 * k + 3]
                if i < 3 + 4 * level:
                    if j[i] in jt:
                        # Same index in group of 4.
                        return True
                    jt.append(j[i])
        return False


    def level_iter(self, r, s=4):
        """
        Generates index tuples to evaluate a general tensor.
        """
        from itertools import permutations
        return permutations(range(r), s)



    #
    # Hybrid version: evaluate at MaMu + tensor
    #

    def clash_mamu_plus_tensor(self, j, level):
        t1 = self.t1
        t2 = self.t2
        jt = []

        counts = 10*[0,]
        found = None
        for i in range(3 + 4 * level):
            x = j[i]
            counts[x] += 1
            if counts[x] >= 5:
                if found == None:
                    found = x
                elif not found == x:
                    return True

        #k = 0
        del jt[:]
        for m in (0,1,2):
            i = t1[m]
            if i < 3 + 4 * level:
                if j[i] in jt:
                    # Same index in group of 4.
                    return True
                jt.append(j[i])
        # Specific to mamu.
        for B in ((0, 1), (2, 3)):  # (4, 5) (6,7)
            if B[0] in jt and B[1] in jt:
                # Same vector in group of 4.
                return True

        del jt[:]
        for m in (0,1,2):
            i = t2[m]
            if i < 3 + 4 * level:
                if j[i] in jt:
                    # Same index in group of 4.
                    return True
                jt.append(j[i])
        # Specific to mamu.
        for B in ((0, 2), (1, 3)):  # (4, 6), (5, 7)
            if B[0] in jt and B[1] in jt:
                # Same vector in group of 4.
                return True

        for k in (0,1,2,3):
            del jt[:]
            for m in (0,1,2,3):
                i = t1[m + 4 * k + 3]
                if i < 3 + 4 * level:
                    if j[i] in jt:
                        # Same index in group of 4.
                        return True
                    jt.append(j[i])
            # Specific to mamu.
            for B in ((0, 1), (2, 3)):  # (4, 5) (6,7)
                if B[0] in jt and B[1] in jt:
                    # Same vector in group of 4.
                    return True

            del jt[:]
            for m in (0,1,2,3):
                i = t2[m + 4 * k + 3]
                if i < 3 + 4 * level:
                    if j[i] in jt:
                        # Same index in group of 4.
                        return True
                    jt.append(j[i])
            # Specific to mamu.
            for B in ((0, 2), (1, 3)): #(4, 6), (5, 7)
                if B[0] in jt and B[1] in jt:
                    # Same vector in group of 4.
                    return True
        return False


if __name__ == "__main__":

    import argparse
    import ast

    parser = argparse.ArgumentParser(description='Find basis')
    parser.add_argument('-basis', action="store_true", help='find basis with perturbations')
    parser.add_argument('-ideal', action="store_true", help='find basis for ideal')
    parser.add_argument('-verify', action="store_true", help='check basis for ideal by straightening')
    parser.add_argument('-verify_symb', action="store_true", help='check basis for ideal by symbolic computation')
    parser.add_argument('-mamu_param', action="store_true", help='evaluate on parametrized 2x2 mamu')
    parser.add_argument('-u', action="store_true", help='run unittests')
    parser.add_argument('-mod', metavar='n', type=int, default=120, help='job granularity')
    parser.add_argument('-rank', metavar='n', type=int, default=6, help='rank')
    parser.add_argument('-single', metavar='i', type=int, default=None, help='single basis element')
    parser.add_argument('-basisfile', metavar='f', type=str, help='file containing basis')
    parser.add_argument('-coeffsfile', metavar='f', type=str, help='file containing ideal basis coeffs')
    parser.add_argument('-shelve', action="store_true", help='use shelve')
    parser.add_argument('-mp_pattern', metavar='p', type=str, help='multi-prolongation pattern')
    args = parser.parse_args()

    if args.u:
        # Another way to run the unittests is by doing python -m unittest poly5555
        runner = unittest.TextTestRunner()
        itersuite = unittest.TestLoader().loadTestsFromTestCase(TestZuiddam)
        runner.run(itersuite)
        exit()

    Z = Zuiddam(mod=args.mod)
    r = args.rank

    if args.basis:
        print "Constructing basis", time.ctime()
        basis = Z.construct_basis()
        print "Done", time.ctime()

    if args.ideal:
        print "Constructing ideal basis", time.ctime()
        if not args.basis:
            print "Reading basis from file."
            if args.basisfile:
                infile = open(args.basisfile, 'r')
                basis = eval(infile.read())
                infile.close()
            else:
                print "No filename given with -basisfile option."
                exit(1)
        kernel_basis_coeffs = Z.construct_ideal_basis(basis,r)
        print "Done", time.ctime()

    if args.verify:
        print "Verifying ideal basis."
        if not args.ideal:
            print "Reading ideal basis coeffs from file."
            if args.coeffsfile:
                infile = open(args.coeffsfile, 'r')
                kernel_basis_coeffs = eval(infile.read())
                infile.close()
            else:
                print "No filename given with -coeffsfile option."
                exit(1)
        if not args.basis:
            print "Reading basis from file."
            if args.basisfile:
                infile = open(args.basisfile, 'r')
                basis = eval(infile.read())
                infile.close()
            else:
                print "No filename given with -basisfile option."
                exit(1)
        if args.mp_pattern:
            mp_pattern = eval(args.mp_pattern)
        else:
            mp_pattern = None
        if args.shelve:
            if not args.single is None:
                print "Single basis element:", args.single
                evaluation = Z.verify_ideal_basis_shelve_single(basis, kernel_basis_coeffs,r,mp_pattern, args.single)
            else:
                evaluation = Z.verify_ideal_basis_shelve(basis, kernel_basis_coeffs,r,mp_pattern)
        else:
            evaluation = Z.verify_ideal_basis(basis, kernel_basis_coeffs,r,mp_pattern)
        print "Done", time.ctime()

    if args.verify_symb:
        print "Verifying ideal basis symbolically."
        if not args.ideal:
            print "Reading ideal basis coeffs from file."
            if args.coeffsfile:
                infile = open(args.coeffsfile, 'r')
                kernel_basis_coeffs = eval(infile.read())
                infile.close()
            else:
                print "No filename given with -coeffsfile option."
                exit(1)
        if not args.basis:
            print "Reading basis from file."
            if args.basisfile:
                infile = open(args.basisfile, 'r')
                basis = eval(infile.read())
                infile.close()
            else:
                print "No filename given with -basisfile option."
                exit(1)
        evaluation = Z.verify_ideal_basis_symbolically(basis, kernel_basis_coeffs, r)
        print "Done", time.ctime()

    if args.mamu_param:
        print "Evaluating on parametrized 2x2 mamu."
        if not args.ideal:
            print "Reading ideal basis coeffs from file."
            if args.coeffsfile:
                infile = open(args.coeffsfile, 'r')
                kernel_basis_coeffs = eval(infile.read())
                infile.close()
            else:
                print "No filename given with -coeffsfile option."
                exit(1)
        if not args.basis:
            print "Reading basis from file."
            if args.basisfile:
                infile = open(args.basisfile, 'r')
                basis = eval(infile.read())
                infile.close()
            else:
                print "No filename given with -basisfile option."
                exit(1)
        evaluation = Z.evaluate_mamu_param(basis, kernel_basis_coeffs)
        print "Done", time.ctime()
