
"""
Tableau straightening for the general linear group

This code is a python port of the straighten function of the package SchurFunctors in Macaulay2, written by Stillman and others, see http://www.math.uiuc.edu/Macaulay2/doc/Macaulay2-1.6/share/Macaulay2/SchurFunctors.m2

"""

import itertools

def normalize(T):
    """
    normalize = method()
    normalize Filling := (T) -> (
         -- returns (c,T'), where c is 0,1 or -1.
         -- T' is T with rows sorted
         -- c=0 for repeats in rows, else {1,-1} is sign of permutation needed to sort
         coeff := 0;
         coeffzero := false;
         T' := apply(T, t -> (
               (c,t') := sortLen t;
               if c < 0 then coeffzero = true;
               coeff = coeff + c;
               t'));
         if coeffzero
        then (0,null)
        else
          (if coeff % 2 == 0 then 1 else -1, new Filling from T')
         )
    """
    coeff = 0
    Tprime = []
    for t in T:
        c, tprime = sortLen(t)
        if c < 0:
            return (0, None)
        coeff += c
        Tprime.append(tprime)

    return ((-1)**coeff, tuple(Tprime))

def sortLen(L):
    """
    sortLen = (L) -> (
         -- L is a list of integers
         -- returned: (s, L')
         -- s is the length of the permutation to place L into order
         -- s will be -1 if L contains duplicate entries
         len := 0;
         s := new MutableList from L;
         n := #s;
         for i from 0 to n-2 do
              for j from 0 to n-i-2 do (
               if s#j === s#(j+1) then return (-1,L);
               if s#j > s#(j+1) then (
                tmp := s#(j+1);
                s#(j+1) = s#j;
                s#j = tmp;
                len = len+1;
                )
               );
         (len, toList s))
    """
    length = 0
    s = list(L[:])
    n = len(s)
    i = 0
    while i < n-1:
        j = 0
        while j < n-i-1:
            if s[j] == s[j+1]:
                return (-1, L)
            if s[j] > s[j+1]:
                tmp = s[j+1]
                s[j+1] = s[j]
                s[j] = tmp
                length += 1
            j += 1
        i += 1
    return (length, tuple(s))

def sortSign(L):
    """
    sortSign = (L) -> (
         (len,L1) := sortLen L;
         (if len =!= -1 then (if len % 2 === 0 then 1 else -1), L1))
    """
    length, L1 = sortLen(L)
    if length == -1:
        return (None, L1)
    else:
        return ((-1)**length, L1)


def isStandard(T):
    """
    isStandard = (T) -> (
     i := #T-2;
     while i >= 0 do (
      a := T#i;
      b := T#(i+1);
      n := #b;
      for j from 0 to n-1 do
        if a#j > b#j then return (i,j);
      i = i-1;
      );
     null
     )
    """

    i = len(T) - 2
    while i >= 0:
        a = T[i]
        b = T[i + 1]
        n = len(b)
        j = 0
        while j < n:
            if a[j] > b[j]:
                return (i,j)
            j += 1
        i -= 1
    return True

def exchange(T, col1, col2, s):
    """
    exchange = (T, col1, col2, s) -> (
         -- s should be a list of positions of T#col1, that will be placed into col2
         -- The returned value is {(coeff, T')}
         -- coeff is 1 or -1.  The length of the list is 0 or 1.
         b := T#col2;
         M := new MutableList from T#col1;
         b = join(apply(#s, i -> (j := s#i; a := M#j; M#j = b#i; a)), drop(b,#s));
         (sgn, M1) := sortSign M;
         (sgnb, b1) := sortSign b;
         if sgn === null or sgnb === null then null else
         (for i from 0 to #T-1 list (
          if i == col1 then M1 else if i == col2 then b1 else T#i
          ), sgn*sgnb)
         )
    """
    b = T[col2]
    M = list(T[col1][:])
    c = []
    for i in range(len(s)):
        j = s[i]
        a = M[j]
        M[j] = b[i]
        c.append(a)
    b = c + list(b[len(s):])
    sgn, M1 = sortSign(M)
    sgnb, b1 = sortSign(b)
    if sgn is None or sgnb is None:
        return None
    else:
        return (tuple(tuple(M1) if i == col1 else (tuple(b1) if i == col2 else T[i]) for i in range(len(T))) , sgn * sgnb)



def towardStandard(T):
    """
    -- writes T as a linear combination of other tableaux T' s.t. T'<T
    -- if T is not standard
    towardStandard = (T) -> (
         x := isStandard T;
         if x === null
           then new HashTable from {T=>1}
           else (
            new HashTable from shuffle(T, x#1+1, x#0, x#0+1)
            )
         )
    """
    x = isStandard(T)
    if x == True:
        return {T: 1}
    else:
        nrows = x[1]+1
        col1 = x[0]
        col2 = x[0]+1
        a = T[col1]
        b = T[col2]
        I = itertools.combinations(range(len(a)), r=nrows)

        output = {}
        for x in I:
            exch = exchange(T, col1, col2, x)
            if not exch == None:
                output[exch[0]] = exch[1]

        return output


def straighten(t, h=None):
    """
    -- Input:
    -- List t: a tableau
    -- MutableHashTable h: a hash table which memoizes recursive calls and
    -- stores the coefficients of the straightening of t into standard tableaux
    straighten = method()
    straighten(List) := t -> (
         t = apply(t, i->sort(i));
         h := new MutableHashTable from {};
         straighten(t, h);
         h#t
         )
    straighten(List, MutableHashTable) := (t, h) -> (
         t = apply(t, i -> sort i);
         if h #? t then return null;
         if isStandard(t) === null then
         h#t = new HashTable from {t => 1};

         firstIter := towardStandard(t);
         H := hashTable({}); -- straightening of t
         for i in keys firstIter do (
          coeff := firstIter#i;
              straighten(i, h);
          temp := {};
          for j in keys h#i do temp = append(temp, (j, coeff * (h#i)#j));
          H = merge(H, hashTable(temp), plus);
              temp = {};
          for j in keys H do if H#j != 0 then temp = append(temp,(j,H#j));
          H = hashTable(temp);
          );
         h#t = H;
         return null;
         )
    """

    if h is None:
        c, t = normalize(t)
        if c == 0:
            return {}
        h = {}
        straighten(t,h)
        return {k: c*v for k,v in h[t].iteritems()}
    else:
        if t in h:
            return True
        if isStandard(t) == True:
            h[t] = {t: 1}
            return True

        firstIter = towardStandard(t)
        H = {}
        for i in firstIter:
            coeff = firstIter[i]

            straighten(i, h)

            for j in h[i]:
                if j in H:
                    x = H[j] + coeff * h[i][j]
                    if x == 0:
                        del H[j]
                    else:
                        H[j] = x
                else:
                    H[j] = coeff * h[i][j]

        h[t] = H
        return True

def straighten3(t1, t2, t3, h=None):
    if h is None:
        c1, t1 = normalize(t1)
        c2, t2 = normalize(t2)
        c3, t3 = normalize(t3)
        h = {}
        if c1*c2*c3 == 0:
            return {}
        straighten3(t1, t2, t3, h)
        return {(k1, k2, k3): c1*c2*c3*v1*v2*v3 for k1,v1 in h[t1].iteritems() for k2,v2 in h[t2].iteritems() for k3,v3 in h[t3].iteritems()}
    else:
        straighten(t1, h)
        straighten(t2, h)
        straighten(t3, h)
        return True

def straightenh(t, h):
    c, t = normalize(t)
    if c == 0:
        return {}
    straighten(t, h)
    return {k: c*v for k,v in h[t].iteritems()}


def straighten3h(t1, t2, t3):
    c1, t1 = normalize(t1)
    c2, t2 = normalize(t2)
    c3, t3 = normalize(t3)
    if c1*c2*c3 == 0:
        return {}
    h1=straighten(t1)
    h2=straighten(t2)
    h3=straighten(t3)

    return {(k1, k2, k3): c1*c2*c3*v1*v2*v3 for k1,v1 in h1.iteritems() for k2,v2 in h2.iteritems() for k3,v3 in h3.iteritems() if v1*v2*v3 != 0}


if __name__ == '__main__':
    import sys
    import argparse

    parser = argparse.ArgumentParser(description='Straighten tableau in GL-module.')
    parser.add_argument('tableau', metavar='t', type=str, help='tableau to straighten, as list of columns')
    args = parser.parse_args()

    t = eval(args.tableau)

    print straighten(t)
